/*
\begin{verbatim}
//////////////////////////////////////////////////////////////////////
//
// Verified Systems International GmbH
// Prof. Dr. Jan Peleska
// Parkstrasse 78 
// D-28209 Bremen
// Germany
// Tel.  : +49 421 344576
// Fax   : +49 421 344586
// e-mail: jp@informatik.uni-bremen.de
//
//--------------------------------------------------------------------
//
// (C) Copyright Verified Systems International GmbH $Date: 2000/09/26 06:11:24 $
//
//--------------------------------------------------------------------
//
// Produkt: RTT - Interface Module
//
//--------------------------------------------------------------------
//
// File Identification: IFM.c
//
// $Header: /lion/home/repository/VVTCVS/CVS/rtt-demo/ifm1/src/rttdemo1.c,v 1.1.1.1 2000/09/26 06:11:24 steenbo Exp $
//
// $Revision: 1.1.1.1 $
//
// First edition by: Jan Peleska
// Last update by $Author: steenbo $
//
//--------------------------------------------------------------------
//
// Description: Source file implementing the 
//              demo target process acting as System under Test 
//
//--------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef _WINDOWS
#include <winsock2.h>
#include <winbase.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>       
#endif

#include "ifm1.h"
#include "rttiflib.h"
#include "rttemlib.h"
#include "rttconflib.h"
#include "rttisetlib.h"
#include "rttoslib.h"
#include "ifm123.a2c.h"

#define CMDSIZE 32
#define _VERBOSE 1
#define RTT_CTP_SIGNAL 32001
#define USE_S2E 1

/* global variables for all modules (scheduler, features ...) */
INT32 buf[BUFFLEN];
INT32 buf_entries;
char byteArrayTX[BUFFLEN];
char byteArrayRX[BUFFLEN];

INT32 f[MAX_EVENTS];
INT32 fold[MAX_EVENTS];

/* the last timetick, sent by RTT. ext_timetick contains the
   network byte order representation of the 64Bit value */
INT64 ext_timetick,timetick;

INT32 b;

rttIset_t *toTestDriverEvents;
rttIset_t *fromSutEvents;

extern t_rttAMConf     aminfo[MAX_AM];
extern t_rttGlobalConf globalinfo;

t_rttAddress          ifmId;

int  sfdsut; /*--- socket descriptor to communicate with SUT ---*/

#include "ifm123.a2c.c"


/*---------------------------------------------------------------------

rawEventsToFeatures: Generate raw events according to the entries in
                     the last packet sent from RTT

INPUT:  -

OUTPUT: -

RETURN: -

NOTE: -

------------------------------------------------------------------------*/
void rawEventsToFeatures() {
  INT32             len,event;
  t_rttCmdType      ctp;
  t_rttDataType     dtp;
  static char       cmdbuf[CMDSIZE];
  static char       rcvbuf[CMDSIZE];
  char              *cmdptr;
  
  /*--- cmddbuf is a fixed size datagram for the communication with
    the SUT. The first eight bytes are used for the timestamp.
    After that the command follows as an ascii-string
    ---*/
  
  len = sizeof(byteArrayRX);
  while ( rttOk ==  
	  rttGetNextCmd(&ctp,&dtp,(void *)&rcvbuf,&len,byteArrayRX) ) {
    /*-------------------------------------------------------------- 
      the received event can either be an abstract event or a signal, that
      has already been mapped by the ccl via rttEvent2Signal 
      in the first case the ctp is RTT_CTP_ABSTRACT_EVENT in the other
      it is the ctp defined in the file event2signal.txt
      ---------------------------------------------------------------*/ 
    switch (ctp) {
    case RTT_CTP_ABSTRACT_EVENT:
      /* process raw-event lists */
      event=(INT32) *rcvbuf;
#ifdef _VERBOSE
      printf("TM %011lli CCL->IFM CTP %i DTP %XH DATA %i \n",
	     timetick,ctp,dtp,event);fflush(NULL);
      fflush(stdout);
#endif
      /*--- here is the solution using an algorithm: ---*/    
      /*--- first send timestamp ---*/
      *((INT64*)cmdbuf)= htonll(timetick);
      cmdptr= cmdbuf + sizeof(INT64);
      switch(rttChanOf(event)) {
      case ctrl:
	switch(rttParmOf(event,1)) {
	case steady:
	  snprintf(cmdptr,sizeof(cmdbuf),
		   "ctrl.steady.%d",rttParmOf(event,2));
	  break;
	case speedUp:
	  snprintf(cmdptr,sizeof(cmdbuf),
		   "ctrl.speedUp.%d",rttParmOf(event,2));
	  break;
	case slowDown:
	  snprintf(cmdptr,sizeof(cmdbuf),
		   "ctrl.slowDown.%d",rttParmOf(event,2));
	  break;
	} /*--- end of switch(rttParmOf(event,1)) ---*/
	break;
      case speed:
	snprintf(cmdptr,sizeof(cmdbuf),"speed.%d",rttParmOf(event,1));
	break;
      case maxSpeedReached:
	snprintf(cmdptr,sizeof(cmdbuf),"maxSpeedReached");
	break;
      case zeroSpeedReached:
	snprintf(cmdptr,sizeof(cmdbuf),"zeroSpeedReached");
	break;
      default:
	printf("Got undefined channel from test engine!\n");
	fflush(stdout);
	return;
      } /*--- end of switch switch(rttChanOf(event)) ---*/ 
      /*--- send command to SUT ---*/
#ifdef _VERBOSE
      printf("TM %011lli IFM->SUT command %s\n", 
	     timetick, cmdptr);
      fflush(stdout);
#endif
      if ( sizeof(cmdbuf) > 
	   send(sfdsut, cmdbuf, sizeof(cmdbuf), 0) ) {
	fprintf(stdout,"error on sending command to SUT");
	fflush(stdout);
	return;
      }
      break; /*--- end of case RTT_CTP_ABSTRACT_EVENT: ---*/
    case RTT_CTP_SIGNAL:
      /*--- this is alredy mapped to a raw-data-signal and can be send
	to the SUT as it is only the timestamp has to be inserted ---*/
      cmdptr= rcvbuf + sizeof(INT64);
#ifdef _VERBOSE
      printf("TM %011lli CCL->IFM CTP %i DTP %XH DATA %s \n",
	     timetick,ctp,dtp,cmdptr);fflush(NULL);
      fflush(stdout);
#endif
      /*--- in this example command must have a fixed size ---*/
      /*--- check that ---*/
      if(len != sizeof(cmdbuf)) {
	fprintf(stderr, "could not map event to signal wrong size %d\n",
		len);
	fflush(stderr);
	return;
      }
      /*--- the first eight byte of the command buffer must be void ---*/
      /*--- here the timestamp is written to ----*/
      *((INT64*)rcvbuf)= timetick;
#ifdef _VERBOSE
      printf("TM %011lli IFM->SUT command %s\n", 
	     timetick, cmdptr);
      fflush(stdout);
#endif
      if ( sizeof(cmdbuf) > 
	   send(sfdsut, rcvbuf, sizeof(cmdbuf), 0) ) {
	fprintf(stdout,"error on sending command to SUT");
	fflush(stdout);
	return;
      }
      break; /*--- end of case RTT_CTP_SIGNAL: ---*/
    default:
        printf("Got undefined command type from test engine!\n");
	fflush(stdout);
	return;
    } /*--- end of switch(ctp) ---*/
  } /*--- end of while ( rttOk ==  rttGetNextCmd....) ---*/
}


/*---------------------------------------------------------------------

extractRawEvents(): Check status arrays for state changes and prepares
                    the data to be sent to CCL

INPUT: -

OUTPUT: -

RETURN: 
number of bytes to send to testdriver or 
-1 if an error occurs

NOTE: -

------------------------------------------------------------------------*/
INT32 extractRawEvents() {
  int len;
  rttIset_t *tmp;
  /*--- nothing to do if no new events happended ---*/
  
  fflush(stdout);
  if(fromSutEvents->card == 0) return 0;
  /*--- clear set that has been send to testdriver ---*/
  
  rttIsetClear(toTestDriverEvents);
  
  /*--- get Lock to change iset-ptr safe ---*/
  rttGetMutexLock(); 
  tmp= fromSutEvents;
  fromSutEvents= toTestDriverEvents;
  toTestDriverEvents= tmp;
  /*--- release lock ---*/
  rttReleaseMutexLock();
  
  if(-1 == (len= rttIsetMakeNetworkBuffer(toTestDriverEvents, buf,0))) {
    fprintf(stderr,"Error converting iset -> nbuf\n");
    fflush(stderr);
    return -1;
  }
  
  return len;
}

/*---------------------------------------------------------------------

processSignalsFromSUT(): Receives Signals from SUT, maps them to events and
                         stores them in iset

INPUT: -

OUTPUT: -

RETURN: -

NOTE: -

------------------------------------------------------------------------*/
void *processSignalsFromSUT() {
  int msgLen, retval;
  t_rttAEvent       event;
  static char       cmdbuf[CMDSIZE];
  char              *cmdptr;
  /*--- cmddbuf is a fixed size datagram for the communication with
    the SUT. The first eight bytes are used for the timestamp.
    After that the command follows as an ascii-string
    ---*/

  while(true) {
    msgLen = recv(sfdsut, cmdbuf, sizeof(cmdbuf), 0);
    /*--- in this example command must have a fixed size ---*/
    if(msgLen != sizeof(cmdbuf)) {
      fprintf(stderr,"error on receiving from SUT\n");
      fflush(stderr);
      printf("press <cr> to exit\n");
      getchar();   
      exit(1);
    }
#ifdef _VERBOSE
    printf("TM %011lli SUT->IFM command %s\n", 
	   timetick, cmdbuf+sizeof(INT64));
    fflush(stdout);
#endif
    /*--- map signal -> event ---*/
    /*--- this could be done with an algorithm or by using
      rttSignal2Event(...) this example uses an algoritm to do the 
      mapping
      ------*/
    /*--- skip timestamp ---*/
    cmdptr= cmdbuf+sizeof(INT64);
    if(strncmp(cmdptr,"accelerate",strlen("accelerate")) == 0) {
      /*--- first accelerate-event is ACCELERATE0 ---*/
      event= ACCELERATE0; 
      /*--- read the parameter of the signal ---*/
      /*--- parameter follows command after a dot ---*/
      /*--- so it beginns at cmdptr+strlen("accelerate")+1 ---*/
      /*--- add this value to event to get event number ----*/
      event+= atoi(cmdptr+strlen("accelerate")+1);
    } else if(strncmp(cmdptr,"brake",strlen("brake")) == 0) {
      /*--- first accelerate-event is ACCELERATE0 ---*/
      event= BRAKE0; 
      /*--- read the parameter of the signal ---*/
      /*--- parameter follows command after a dot ---*/
      /*--- so it beginns at cmdptr+strlen("brake")+1 ---*/
      /*--- add this value to event to get event number ----*/
      event+= atoi(cmdptr+strlen("brake")+1);
    } else {
      fprintf(stderr,"unknown command from SUT %s\n",cmdptr);
      fflush(stderr);
      continue;
    }

    /*--- get lock because fromSutEvents may be changed by
      extractRawEvents() ---*/
    rttGetMutexLock();
    retval= rttIsetInsertElem(event,fromSutEvents);
    /*--- release lock ---*/
    rttReleaseMutexLock();
    if(0 != retval) {
      fprintf(stdout,"could not insert event %d to iset\n",event);
      if(retval == 1) printf("element already exists\n");
      fflush(stdout);
      continue;
    }
  } /*--- end of while(true) ---*/
}

/*---------------------------------------------------------------------

printUsage(char *fname): prints the usage of the interface module 

INPUT: char *fname: the name of the executable

OUTPUT: -

RETURN: -

NOTE: -

------------------------------------------------------------------------*/

void printUsage(char *fname) {
  char *ifmname;
  char separator;
  
#ifdef _WINDOWS
  separator = '\\';
#else
  separator = '/';
#endif
    
  ifmname = strrchr(fname, separator);
  if (ifmname==NULL) ifmname = fname;
  else ifmname++;
  
  printf("RT-Tester Interface Module\n");
  printf("==========================\n\n");
  printf("Usage: %s <IFMID> <RTTSNDADR> <RTTRCVADR> <MAXEVENTS>\n", ifmname);
  printf("       %s <IFMID> <CONFIGFILE>\n", ifmname);
  printf("       <IFMID>      Id of Interface Module.\n");
  printf("       <RTTSNDADR>  IP-address and port of RTTSNDCCL.\n");
  printf("       <RTTRCVADR>  IP-address and port of RTTRCVCCL.\n");
  printf("       <CONFIGFILE> path to RTT-Configfile.\n");
#ifdef _WINDOWS
  Sleep(5000);
#else
  printf("press <cr> to exit\n");
  getchar();
#endif
  exit(1);
}


/*---------------------------------------------------------------------

main(): The main function of this interface module 

INPUT: -

OUTPUT: -

RETURN: -

NOTE: -

------------------------------------------------------------------------*/

int main(int argc, char **argv) {

  t_rttReturncodes  retcode;

  INT32             len;
  t_rttCmdType      ctp;
  t_rttDataType     dtp;
  struct  sockaddr_in sutaddr;           /* socketaddress of SUT */
  int               connected = 1;
  
  char             *rttsendadr, *rttrecvadr;
  char             *configFileName = NULL;
  char *strSutinetadr= "127.0.0.1";
  char *strSutport= "12700";
  
  int maxevent;
#ifdef _WINDOWS
  WSADATA wsadatasut;
  WSAStartup (WSA_Version, &wsadatasut);
#endif
  
#ifdef _SOL
  bzero((char *)&sutaddr,sizeof(sutaddr));
#endif
  
#ifdef _LINUX
  bzero((char *)&sutaddr,sizeof(sutaddr));
#endif


  /* Checks the arguments given to this IFM */
  if (argc==4) {
	ifmId = atoi(argv[1]);
	rttsendadr = strdup(argv[2]);
	rttrecvadr = strdup(argv[3]);
	maxevent= NO_OF_EVENTS-1;
  } else if (argc==3) {
    ifmId = atoi(argv[1]);
    if(ifmId < 1 || ifmId > MAX_AM) {
      fprintf(stderr,"Invalid ifmId %s\n", argv[1]);
      fflush(stderr);
      printf("press <cr> to exit\n");
      getchar();
      exit(1);
    } 
    configFileName = argv[2];
    if((retcode= rttConfInit(configFileName)) != rttOk) {
      fprintf(stderr,"Error reading configuration files %s\n", configFileName);
      fflush(stderr);
      printf("press <cr> to exit\n");
      getchar();
      exit(retcode);
    } 
    if(!globalinfo.sndccl) {
      fprintf(stderr,"sndccl-Adr not specified in configuration file\n");
      fflush(stderr);      
      printf("press <cr> to exit\n");
      getchar();
      exit(1);
    }
    if(!globalinfo.rcvccl) {
      fprintf(stderr,"rcvccl-Adr not specified in configuration file\n");
      fflush(stderr);
      printf("press <cr> to exit\n");
      getchar();
      exit(1);
    }
    rttsendadr= globalinfo.sndccl;
    rttrecvadr= globalinfo.rcvccl;
    maxevent= NO_OF_EVENTS-1;
  } else {
	printUsage(argv[0]);
  }

  printf("RT-Tester Interface Module %3d\n", ifmId);
  printf("==============================\n\n");
  
  /* Do initialization of IFM here */
  /*--- initialize EventSet to send to testdriver ---*/
  toTestDriverEvents= rttIsetCreate(maxevent);
  fromSutEvents= rttIsetCreate(maxevent);
  if(!toTestDriverEvents || !fromSutEvents) {
    fprintf(stderr,"creating isets failed maxevent: %d\n",maxevent);
    fflush(stderr);
    printf("press <cr> to exit\n");
    getchar();
    exit(1);
  }
  /* connect to CCL */
  printf("Connecting to CCL.\n");
  printf("  rttsndccl at %s\n", rttsendadr);
  printf("  rttrcvccl at %s\n", rttrecvadr);
  fflush(stdout);
  retcode = rttConnectToCCL(ifmId, rttsendadr, rttrecvadr);
  if (retcode!=rttOk) {
    printf("Connect to test engine failed!\n");
    fflush(stdout);
    connected = 0;
#ifdef _WINDOWS
    Sleep(5000);
#else
    printf("press <cr> to exit\n");
    getchar();
#endif
    exit (5);
  }
#ifdef _VERBOSE
  printf("connected with CCL\n");
  fflush(stdout);
#endif
  /*--- connect to SUT ---*/
  /*--- set server socket address for read ---*/
  sutaddr.sin_family = AF_INET;
  sutaddr.sin_addr.s_addr = inet_addr(strSutinetadr);
#ifdef _WINDOWS
  sutaddr.sin_port = htons((u_short)atoi(strSutport));
#else
  sutaddr.sin_port = htons(atoi(strSutport));
#endif
  if ( -1 == (sfdsut = socket( AF_INET, SOCK_STREAM, 0 )) ) {
    printf("Error when creating socket for send \n");
    fflush(stdout);
    printf("press <cr> to exit\n");
    getchar();
    exit(1);
  }
    
  if ( -1 == (connect(sfdsut,(struct sockaddr *)&sutaddr,sizeof(sutaddr) )) ) {
    printf("Error on connect with SUT \n");
    fflush(stdout);
    printf("press <cr> to exit\n");
    getchar();   
    exit(1);
  }
#ifdef _VERBOSE
  printf("connected with SUT\n");
  fflush(stdout);
#endif
    
  /*--- create a Thread for processing signals from SUT ---*/
  
  if(0 != rttActivateThread (processSignalsFromSUT, NULL)) {
    fprintf(stderr, "Error on creating  processSignalsFromSUT\n");
    fflush(stderr); 
    printf("press <cr> to exit\n");
    getchar();   
    exit(1);
  } 
#ifdef _VERY_VERBOSE
  printf("Thread created\n");
  fflush(stdout);
#endif
    

  /* Use this to wait for all Interface Modules */

  /*--- get data from test engine ---*/
  retcode = rttResetByteArray(BUFFLEN,byteArrayRX);
  retcode = rttResetByteArray(BUFFLEN,byteArrayTX);
  rttReceiveFromTestDriver(rttWait,byteArrayRX); /* wait for tick (and data) */
  
  do { 

    /*--- send data from last step to test engine ---*/
    /*	retcode = rttSendToTestDriver(byteArrayTX);
    if ((connected == 1) && (rttOk != retcode)) {
      printf("IFM stopped from RT Tester.\n");
      if (retcode == -8) {
	// Here we have to close everything.
	exit(0);
      }
      }*/

    /* evaluate byteArray entries:
       - extract time stamp */
    len=sizeof(INT64);
    rttGetNextCmd(&ctp,&dtp,(void *)&ext_timetick,&len,byteArrayRX); 
    if (ctp==RTT_CTP_TIMETICK) {
      timetick=ext_timetick;
#ifdef _VERY_VERBOSE		
      printf("TimeTick: %011lli\n",timetick);
      fflush(stdout);
#endif
    }  
    else {
      /*
	printf("Error: First entry is not a time stamp!\n");
      */
    } 
    
    /* extract raw events and generate new raw Events for the test
       engine in byteArrayTX */
    rawEventsToFeatures();
    len= extractRawEvents(); 
    
    if(len == -1) {
      fprintf(stderr,"error extracting rawevents \n");
      fflush(stderr);
    } else if (len > 0) {
      
      retcode = rttResetByteArray(BUFFLEN,byteArrayTX);

      /* copy timetick from current packet */
      retcode = rttPutCmd(RTT_CTP_TIMETICK, 0, (char *)&ext_timetick, 
			  sizeof(INT64), byteArrayTX);
    
      /* put the generated array into the send buffer */ 
    
      retcode = rttPutCmd(RTT_CTP_ABSTRACT_EVENT_ISET,
			  0 , (char *)buf, len, byteArrayTX);
      /*--- send data to test engine ---*/
#ifdef _VERBOSE
      printf("TM %011lli IFM->CCL: Sending events  ", timetick);
      rttIsetDump(stdout,toTestDriverEvents);
      fflush(stdout);
#endif
      retcode = rttSendToTestDriver(byteArrayTX);
    }
    

    /*--- get data from test engine ---*/
    retcode = rttResetByteArray(BUFFLEN,byteArrayRX);

    rttReceiveFromTestDriver(rttWait,byteArrayRX); // wait for tick (and data)
    
  } while (1); 

  return 0;
}

/*//////////////////////////////////////////////////////////////
//
// Change History:
//
// $Log: rttdemo1.c,v $
// Revision 1.1.1.1  2000/09/26 06:11:24  steenbo
// initial.
//
// Revision 1.4  2000/08/21 07:45:46  steenbo
// byteorder done by rttputcmd/rttgetcmd
//
// Revision 1.3  2000/08/17 07:45:58  steenbo
// before seprating two IFMs, ifm2 uses event2signal ifm1 not
//
// Revision 1.2  2000/08/16 13:53:33  steenbo
// correct implementation of using event2signal mapping
//
// Revision 1.1  2000/08/14 15:32:38  steenbo
// Initial revision
//
// Revision 3.2  2000/07/11 19:23:08  jp
// Release Baseline 3.2
//
// Revision 1.4  2000/04/28 13:18:10  dahlweid
// The variable len in rawEventsToFeatures() was not used properly.
//
// Revision 1.3  2000/04/19 05:04:50  dahlweid
// Changed RTT_CTP_RAW_EVENT_LIST to RTT_CTP_ABSTRACT_EVENT_LIST.
//
// Revision 1.2  2000/03/31 13:48:09  dahlweid
// Several additions to template
//
// Revision 1.1  2000/03/30 10:23:01  dahlweid
// - Added new templates directory to rt-tester hierarchy and included it
// in the build process.
// - Added a template for interface modules to templates directory. This
// IFM is ready to compile under Windows and Unix.
//
//
///////////////////////////////////////////////////////////////
\end{verbatim}
*/
