/*
\begin{verbatim}
//////////////////////////////////////////////////////////////////////
//
// Verified Systems International GmbH
// Parkstrasse 78 
// D-28209 Bremen
// Germany
// Tel.  : +49 421 344576
// Fax   : +49 421 344586
// e-mail: info@verified.de
//
//--------------------------------------------------------------------
//
// (C) Copyright Verified Systems International GmbH $Date: 2000/09/26 06:11:24 $
//
//--------------------------------------------------------------------
//
// Produkt: RTT - DEMO APPLICATION
//
//--------------------------------------------------------------------
//
// File Identification: rttsut.c
//
// $Header: /lion/home/repository/VVTCVS/CVS/rtt-demo/sut/src/rttsut.c,v 1.1.1.1 2000/09/26 06:11:24 steenbo Exp $
//
// $Revision: 1.1.1.1 $
//
// First edition by: Jan Peleska
// Last update by $Author: steenbo $
//
//--------------------------------------------------------------------
//
// Description: Source file implementing a sample apllication 
//              acting as System under Test 
//
--------------------------------------------------------------------*/


const char rttsutrcsid[] = "$Id: rttsut.c,v 1.1.1.1 2000/09/26 06:11:24 steenbo Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h> 
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>


#ifndef INT64
#define INT64 long long 
#endif
#ifndef INT32
#define INT32 long 
#endif

#define MAXSPEED 50
#define MAXSPEEDDURATION 5000

/*--- controller statespace ----*/
INT64             timetick;    /*--- current time ---*/
INT32             thisSpeed;   /*--- current speed ----*/
INT32             waitCtr; /*--- time for emergency brake ---*/
/*--- flag showing that max speed has been reached ---*/
INT32             maxSpeedReached; 
/*--- flag showing that zero speed has been reached ---*/
INT32             zeroSpeedReached; 


/*---------------------------------------------------------------------

ntohll : network to host byte order for INT64

INPUT:
 val: value to be changed from network byte order to host byte order

OUTPUT: -

RETURN: the given value in host byte order

NOTE: -

------------------------------------------------------------------------*/
INT64 ntohll (INT64 val) {
  INT64 retval;
  unsigned long *valp, *retp;

  valp = (unsigned long *)&val;
  retp = (unsigned long *)&retval;

  *retp = ntohl(*(valp+1));
  *(retp+1) = ntohl(*valp);

  return retval;
}

/*---------------------------------------------------------------------

htonll :  host to network byte order for INT64

INPUT:
 val: value to be changed from network byte order to host byte order

OUTPUT: -

RETURN: the given value in host byte order

NOTE: -

------------------------------------------------------------------------*/
INT64 htonll (INT64 val) {
  INT64 retval;
  unsigned long *valp, *retp;

  valp = (unsigned long *)&val;
  retp = (unsigned long *)&retval;

  *retp = htonl(*(valp+1));
  *(retp+1) = htonl(*valp);

  return retval;
}


/******************************************************************
     Controller functions
*******************************************************************/

/*---------------------------------------------------------------------

accelerate :  processes a user command to accelerate /brake  
              gets a signed parameter, if this is a negative value
              it shows a slowDown-Request a SpeedUp, steady-Request resp.
	      if greater then or equal 0.
	      the function processes a control command which is
	      given as char * to the calling function

INPUT:
 val:    value to slow down  or speed up 
 cmdbuf: pointer to command buffer to process

OUTPUT: 
 cmdbuf: if return value is 1 cmd contains a command to send 
         to the engine (in this case the IFM)

RETURN:  
-1 if an error occurs 
0  if no command has been processed
1  if command is available

NOTE: -

------------------------------------------------------------------------*/

int accelerate(int cmdval, char *cmdbuf) {
  char *bufptr;
  if(!cmdbuf) return -1;
  bzero((void *) cmdbuf,32);
  /*--- leave timestamp ---*/
  bufptr= cmdbuf + sizeof(INT64);
  if(cmdval>0 && thisSpeed < MAXSPEED) {
    sprintf(bufptr,"accelerate.%d",cmdval);
    return 1;
  } else if (cmdval == 0) { 
    /*--- here is an error! the command brake.0 should only
      be generated if  0 < thisSpeed ---*/
    sprintf(bufptr,"brake.%d",cmdval);
    return 1;
  } else if (cmdval < 0 && 0 < thisSpeed) { 
    sprintf(bufptr,"brake.%d",cmdval*-1);
    return 1;
  }
  return 0;
}

/*---------------------------------------------------------------------

updateSpeed :  the current speed is updated. 
               if the maximum speed is reached, this is noticed in the
               flag maxSpeedReached. In waitCtr the time for an emergency-
	       break through the controller is stored. this is done if the 
               user does not slow down.
	       
	       if speed is under max speed the maxSpeedReached-flag is cleared.
	       

INPUT:
 val:    value to slow down  or speed up 

OUTPUT: 
-
RETURN:  
-
NOTE: -

------------------------------------------------------------------------*/

void updateSpeed(int val) {
  thisSpeed= val;
  if ( !maxSpeedReached && thisSpeed == MAXSPEED ) {
    maxSpeedReached = 1;
    waitCtr = timetick + MAXSPEEDDURATION;
  } 
  if ( thisSpeed < MAXSPEED )
    maxSpeedReached = 0;
}

/*---------------------------------------------------------------------

checkSpeed :   checks the current speed:
	       if maximum speed has been reached before, the necessarity for
               an ememergency brake is checked.

               this function processes a control command which is
	       given as char * to the calling function

INPUT:
 cmdbuf: pointer to command buffer to process

OUTPUT: 
 cmdbuf: if return value is 1 cmdbuf contains a command to send 
         to the engine (in this case the IFM)

RETURN:  
-1 if an error occurs 
0  if no command has been processed
1  if command is available

NOTE: -

------------------------------------------------------------------------*/

int checkSpeed(char *cmdbuf) {
  char *bufptr;
  bzero((void *) cmdbuf,32);
  /*---- emergency brake ? ----*/
  if ( waitCtr && (maxSpeedReached && waitCtr <= timetick) ) {
    /*---- emergency brake ! ---*/
    if(!cmdbuf) return -1;
    /*--- leave timestamp ---*/
    bufptr= cmdbuf + sizeof(INT64);
    sprintf(bufptr,"brake.5");
    waitCtr= 0;
    return 1;
  }
  return 0;
}

/*---------------------------------------------------------------
  main program
------------------------------------------------------------------*/


void main(int argc, char *argv[]) {

  int  sfd, sfdx;        /*--- socket descriptor to communicate with IFM ---*/
  struct  sockaddr_in saddr, cliaddr; 
  int clilen;
  INT32             len;
  int               accelerateSign;


  char              cmdbuf[32];
  char              *tokenptr;
  char              *cmdptr;

  INT32             cmdval;

  INT32             toEngine;

  int optval = 1;
  
  
#ifdef _WIN32
  WSADATA wsadatasut;
  WSAStartup (WSA_Version, &wsadatasut);
  clilen= 0;
#endif
  
#ifndef _WIN32
  bzero((char *)&saddr,sizeof(saddr));
  clilen = sizeof(struct sockaddr);
#endif
  
/*--- accept connection from IFM ---*/
  /*--- set server socket address for read ---*/
  saddr.sin_family = AF_INET;
  saddr.sin_addr.s_addr = htonl(INADDR_ANY);

  if (argc >= 2) {
#ifdef _WIN32
    saddr.sin_port = htons((u_short)atoi(argv[1]));
#else
    saddr.sin_port = htons(atoi(argv[1]));
#endif
  } else {
    saddr.sin_port = htons(12700);
  }
  
  
  if ( -1 == (sfd = socket( AF_INET, SOCK_STREAM, 0 )) ) {
    fprintf(stderr,"Error when creating socket\n");
    fflush(stderr);
    exit(1);
  }
  
  if ( -1 == setsockopt(sfd,SOL_SOCKET,SO_REUSEADDR,
			&optval,sizeof(optval)) ) {
    fprintf(stderr, 
	    "Error when setting socket option SO_REUSEADDR\n");
    fflush(stderr);
  }  
  if ( -1 == setsockopt(sfd,SOL_TCP,TCP_NODELAY,
			&optval,sizeof(optval)) ) {
    fprintf(stderr, 
	    "Error when setting socket option TCP_NODELAY\n");
    fflush(stderr);
  }
  if ( -1 == bind(sfd,(struct sockaddr *)&saddr,sizeof(saddr)) ) {
    fprintf(stderr, "Error on bind\n");
    fflush(stderr);
    exit(1);
  }
 
  listen(sfd,5);
#ifdef _VERBOSE
  printf("SUT: waiting for engine to connect\n");
  fflush(stdout);
#endif

#ifdef _WIN32
  sfdx = accept(sfd, (struct sockaddr *) &cliaddr, (int FAR*) clilen);
#endif

#ifndef _WIN32
  sfdx = accept(sfd, (struct sockaddr *) &cliaddr, &clilen);
#endif

#ifdef _VERBOSE
  printf("SUT: connected with IFM\n");
  fflush(stdout);
#endif  
  do { 

    /*--- get command from ifm ---*/
    len = recv(sfdx, cmdbuf , sizeof(cmdbuf) ,0);
    if(len != sizeof(cmdbuf)) {
      fprintf(stderr,"Error on receiving Data from IFM, len is: %ld\n",len);
      fflush(stderr);
      exit(1);
    }

    timetick=  ntohll(*(INT64*) cmdbuf);
    cmdptr= cmdbuf + sizeof(INT64);
#ifdef _VERBOSE
    printf("TM %011lli IFM->SUT %s\n",
	   timetick, cmdptr); 
    fflush(stdout);
#endif
    tokenptr= strtok(cmdptr,".");
    
    if(tokenptr) {
      if(strcmp(tokenptr,"ctrl") == 0) {
	if(!(tokenptr= strtok(NULL,"."))) {
	  fprintf(stderr, "invalid command from SUT: %s\n", tokenptr);
	  fflush(stderr);
	  continue;
	}
	if(strcmp(tokenptr,"steady") == 0) {
	  accelerateSign = 0;
	} else if(strcmp(tokenptr,"speedUp") == 0) {
	  accelerateSign = 1;
	} else if(strcmp(tokenptr,"slowDown") == 0) {
	  accelerateSign = -1;
	} else {
	  fprintf(stderr, "invalid command from SUT: ctrl.%s\n", tokenptr);
	  fflush(stderr);
	  continue;
	}
	/*--- read value ----*/
	if(!(tokenptr= strtok(NULL,"."))) {
	    fprintf(stderr, "invalid command from SUT: %s\n", tokenptr);
	    fflush(stderr);
	    continue;
	}
	cmdval= atoi(tokenptr);
	/*--- if command steady is given cmdvalue must be 0
	  it must be <> 0 else ---*/
	if(!((!cmdval && !accelerateSign) || (cmdval && accelerateSign))  ) {
	  fprintf(stderr, "invalid command from SUT %s\n", 
		  tokenptr);
	  fflush(stderr);
	  continue;
	} 
	if(1==accelerate(accelerateSign*cmdval,cmdbuf)) toEngine= 1;
      } else if (strcmp(tokenptr,"speed") == 0) {
	if(!(tokenptr= strtok(NULL,"."))) {
	  fprintf(stderr, "invalid command from SUT: %s\n", tokenptr);
	  fflush(stderr);
	  continue;
	}
	cmdval= atoi(tokenptr);
	updateSpeed(cmdval);
      } else if (strcmp(tokenptr,"maxSpeedReached") == 0) {
	/*--- ignore this information ---*/
      } else if (strcmp(tokenptr,"zeroSpeedReached") == 0) {
	/*--- ignore this information ---*/
      } else {
	fprintf(stderr, "invalid command from SUT: %s\n", tokenptr);
	fflush(stderr);
	continue;
      }
    } else {
      fprintf(stderr, "invalid command-format from SUT: %s\n", tokenptr);
      fflush(stderr);
      continue;
    }
    
    if(toEngine) {
      /*--- send command to engine (IFM) ----*/
      /*--- insert timetick ----*/
      *(INT64*)cmdbuf= htonll(timetick);
#ifdef _VERBOSE
      printf("TM %011lli SUT->IFM %s\n", 
	     timetick,((char*)cmdbuf)+sizeof(INT64));
      fflush(stdout);
#endif
      if ( sizeof(cmdbuf) != 
	   send(sfdx, cmdbuf, sizeof(cmdbuf), 0) ) {
	fprintf(stdout,"error on sending command to Engine\n");
	fflush(stdout);
      }
      toEngine=0;
    }
    /*--- check the current speed ---*/
    if(1==checkSpeed(cmdbuf)) toEngine= 1; 
    
    if(toEngine) {
      /*--- send command to engine (IFM) ----*/
      *(INT64*)cmdbuf= htonll(timetick);
#ifdef _VERBOSE
      printf("TM %011lli SUT->IFM %s\n", 
	     timetick,((char*)cmdbuf)+sizeof(INT64));
      fflush(stdout);
#endif
      if ( sizeof(cmdbuf) > 
	   send(sfdx, cmdbuf, sizeof(cmdbuf), 0) ) {
	fprintf(stdout,"error on sending command to Engine\n");
	fflush(stdout);
      }
      toEngine=0;
    }
    
  } while (1);
    
}
/*//////////////////////////////////////////////////////////////
//
// Change History:
//
// $Log: rttsut.c,v $
// Revision 1.1.1.1  2000/09/26 06:11:24  steenbo
// initial.
//
// Revision 2.8  1999/12/13 17:37:54  steenbo
// Headers and RCV-Vars inserted
//
// Revision 2.7  1999/12/07 06:03:06  jp
// .
//
// Revision 2.6  1999/12/07 06:01:51  jp
// .
//
// Revision 2.5  1999/12/01 13:29:46  jp
// new revision
//
// Revision 1.1.1.1  1999/12/01 07:52:49  jp
// initial
//
// Revision 1.2  1998/12/02 10:26:05  jp
// .
//
// Revision 1.1  1998/10/28 10:51:23  jp
// new demo target system
//
//
//
///////////////////////////////////////////////////////////////
\end{verbatim}
*/
